import re
import string

number_and_letter_together_at_end = re.compile(r"([a-zA-Z]{1,2}[0-9]+)")
number_and_letter_together_at_begin = re.compile(r"([0-9]+[a-zA-Z]{1,2})")
numbers_with_punct = re.compile(r"([0-9]+.?[0-9]+)")
punct_with_numbers = re.compile(r"(\^|\$|!|@|#|%|¨|&|\*|\(|\))[0-9]+")
string_is_date = re.compile(r"((^[0-9]{1,2}.)?([0-9]{1,2}).([0-9]{2,4})$)")


def is_date(token) -> bool:
    if string_is_date.match(token.text):
        return True
    return False


def is_letter_without_meaning(token) -> bool:
    if (
        len(token.text) == 1
        and token.text in "bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ"
    ):
        return True
    return False


def have_letter_and_number_together(token) -> bool:
    if number_and_letter_together_at_end.match(
        token.text
    ) or number_and_letter_together_at_begin.match(token.text):
        return True
    return False


def have_only_punct(token) -> bool:
    if len({c for c in token.text} - {p for p in string.punctuation}) == 0:
        return True
    return False


def has_numbers_with_punct(token) -> bool:
    if numbers_with_punct.match(token.text):
        return True
    return False


def has_punct_with_numbers(token) -> bool:
    if punct_with_numbers.match(token.text):
        return True
    return False



from enum import Enum


class Grade(str, Enum):
    VERY_EASY: str = "very easy"
    EASY: str = "easy"
    FAIRLY_DIFFICULT: str = "fairly difficult"
    VERY_DIFFICULT: str = "very difficult"
    UNKNOWN: str = "unknown"

    def __str__(self) -> str:
        return str.__str__(self)



from ._logger import logger
from .document import Document
from .score import Score


def flesch_index_pt_br(document: Document) -> Score:
    """Readability formulas applied to textbooks in brazilian portuguese.

    Reference:
    - http://www.nilc.icmc.usp.br/nilc/download/Reltec28.pdf
    """
    n_words = len(document)
    try:
        index = int(
            248.835
            - (1.015 * n_words / len(document.sentences))
            - (84.6 * document.syllables / n_words)
        )
    except ZeroDivisionError:
        index = -1
    logger.debug(f"Index computed: {index}")
    if index > 100:
        index = 100
    logger.debug(f"Index adjusted: {index}")
    return Score(index)



import pytest
import spacy

from text_grade import Document

TEXT = """
PREÂMBULO

Nós, representantes do povo brasileiro, reunidos em Assembléia Nacional Constituinte para instituir um Estado Democrático, destinado a assegurar o exercício dos direitos sociais e individuais, a liberdade, a segurança, o bem-estar, o desenvolvimento, a igualdade e a justiça como valores supremos de uma sociedade fraterna, pluralista e sem preconceitos, fundada na harmonia social e comprometida, na ordem interna e internacional, com a solução pacífica das controvérsias, promulgamos, sob a proteção de Deus, a seguinte CONSTITUIÇÃO DA REPÚBLICA FEDERATIVA DO BRASIL.
"""


@pytest.fixture(scope="module")
def nlp():
    nlp = spacy.load("pt_core_news_sm")
    return nlp


@pytest.fixture(scope="module")
def document(nlp):
    return Document(nlp(TEXT))



from text_grade import Grade, formulas


def test_flesch_index_pt_br(document):
    score = formulas.flesch_index_pt_br(document)
    assert score.value <= 10
    assert score.grade == Grade.VERY_DIFFICULT



from typing import Iterator

import pandas as pd
import seaborn as sns

from .document import Document


def characters_x_words(documents: Iterator[Document]):
    return sns.scatterplot(
        data=pd.concat([doc.to_df() for doc in documents]),
        x="n_words",
        y="n_characters",
    )


def characters_boxplot(documents: Iterator[Document]):
    return sns.boxplot(
        data=pd.DataFrame(
            [doc.characters() for doc in documents], columns=["characters"]
        )
    )


def words_boxplot(documents: Iterator[Document]):
    return sns.boxplot(
        data=pd.DataFrame([doc.characters() for doc in documents], columns=["words"])
    )


def sentences_boxplot(documents: Iterator[Document]):
    return sns.boxplot(
        data=pd.DataFrame(
            [doc.characters() for doc in documents], columns=["sentences"]
        )
    )


def sentences_x_words(documents: Iterator[Document]):
    return sns.relplot(
        data=pd.concat([doc.to_df() for doc in documents]), x="n_sentences", y="n_words"
    )


def words_x_characters(documents: Iterator[Document]):
    return sns.relplot(
        data=pd.concat([doc.to_df() for doc in documents]),
        x="n_words",
        y="n_characters",
    )


def words_x_sentences(documents: Iterator[Document]):
    return sns.scatterplot(
        data=pd.concat([doc.to_df() for doc in documents]), x="n_sentences", y="n_words"
    )


def syllables_x_words(documents: Iterator[Document]):
    return sns.scatterplot(
        data=pd.concat([doc.to_df() for doc in documents]), x="n_words", y="syllables"
    )


def unique_words_distribution(documents: Iterator[Document]):
    pass


def score_count(documents: Iterator[Document], formula):
    return sns.countplot(
        data=pd.DataFrame(
            [
                (score.value, str(score.grade))
                for score in [formula(doc) for doc in documents]
            ],
            columns=["value", "score"],
        ),
        x="score",
    )


def score_stripplot(documents: Iterator[Document], formula):
    return sns.stripplot(
        data=pd.DataFrame(
            [
                (score.value, str(score.grade))
                for score in [formula(doc) for doc in documents]
            ],
            columns=["value", "score"],
        ),
        x="score",
        y="value",
    )


def score_boxplot(documents: Iterator[Document], formula):
    return sns.boxplot(
        data=pd.DataFrame(
            [
                (score.value, str(score.grade))
                for score in [formula(doc) for doc in documents]
            ],
            columns=["value", "score"],
        ),
        x="score",
        y="value",
    )



import pytest

from text_grade import filter


@pytest.mark.parametrize(
    "string, expected",
    [
        ("20/12/2022", True),
        ("20/12/22", True),
        ("20/12/2022", True),
        ("20.12.2022", True),
        ("20 12 2022", False),
        ("20-12-2022", True),
        ("12/2022", True),
        ("12.2022", True),
        ("12 2022", False),
    ],
)
def test_is_date(nlp, string, expected):
    doc = nlp(string)
    assert filter.is_date(doc[0]) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("m", True),
        ("mn", False),
        ("a", False),
        ("e", False),
        ("I", False),
        ("o", False),
        ("U", False),
        ("z", True),
        ("k", True),
        ("kk", False),
    ],
)
def test_is_letter_without_meaning(nlp, string, expected):
    doc = nlp(string)
    assert filter.is_letter_without_meaning(doc[0]) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("m22", True),
        # ("22m", True),
        ("mn30303", True),
        ("30303mn", True),
        ("asdfasa", False),
        ("eadsfasf", False),
        ("I(I", False),
        ("999", False),
    ],
)
def test_have_letter_and_number_together(nlp, string, expected):
    doc = nlp(string)
    assert filter.have_letter_and_number_together(doc[0]) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("alex", False),
        ("alex!", False),
        ("!", True),
        (".!?", True),
        ("@@", True),
        (",.", True),
        ("2,2", False),
        (".^}]", True),
    ],
)
def test_have_only_punct(nlp, string, expected):
    doc = nlp(string)
    assert filter.have_only_punct(doc[0]) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("2-22", True),
        ("2.22", True),
        ("0000.0000", True),
        ("2:2", True),
        ("2?2", True),
        ("222", False),
        ("22 22", False),
        ("22.22.22", True),
    ],
)
def test_has_number_with_punct(nlp, string, expected):
    doc = nlp(string)
    assert filter.has_numbers_with_punct(doc[0])


# @pytest.mark.parametrize(
#     "string, expected",
#     [
#         ("2-22", False),
#         ("2.22", False),
#         ("0000.0000", False),
#         ("2:2", False),
#         ("2?2", False),
#         ("222", False),
#         ("^22", True),
#         ("!22", True),
#         ("$2222", True),
#     ],
# )
# def test_has_punct_with_numbers(nlp, string, expected):
#     doc = nlp(string)
#     assert filter.has_punct_with_numbers(doc[0]) == expected



from .grade import Grade


class Score:
    __slots__ = ("_value", "_grade")

    def __init__(self, value: int) -> None:
        self._value = value
        self._grade = self._calculate_grade(value)

    @property
    def value(self) -> int:
        return self._value

    @property
    def grade(self) -> Grade:
        return self._grade

    @property
    def grade_description(self) -> str:
        return str(self._grade)

    def _calculate_grade(self, value: int) -> Grade:
        if value > 75:
            return Grade.VERY_EASY
        elif value <= 75 and value > 50:
            return Grade.EASY
        elif value <= 50 and value > 25:
            return Grade.FAIRLY_DIFFICULT
        elif value <= 25 and value >= 0:
            return Grade.VERY_DIFFICULT
        else:
            return Grade.UNKNOWN

    def __str__(self) -> str:
        return self._grade

    def __repr__(self) -> str:
        return f"Score(value={self._value}, grade='{self.grade_description}')"



from functools import cached_property
from typing import Dict, Generator, List, Optional, Set

import pandas as pd
from pyphen import Pyphen
from spacy.tokens.doc import Doc
from spacy.tokens.span import Span
from spacy.tokens.token import Token

from ._logger import logger
from .filter import (
    has_numbers_with_punct,
    has_punct_with_numbers,
    have_letter_and_number_together,
    have_only_punct,
    is_date,
    is_letter_without_meaning,
)


class Document:
    def __init__(self, doc: Doc) -> None:
        self._doc = doc
        self.tokens: List[Token] = [
            token for token in doc if self._filter_token(token) is not None
        ]
        logger.debug(f"Document language: '{doc.lang_}'")
        self._pyphen = Pyphen(lang=doc.lang_)

    def _filter_token(self, token: Token) -> Optional[Token]:
        if (
            token.is_bracket
            or token.is_currency
            or token.is_digit
            or token.is_punct
            or token.is_quote
            or token.is_space
            or token.like_email
            or token.like_url
            or token.like_num
            or is_date(token)
            or is_letter_without_meaning(token)
            or have_letter_and_number_together(token)
            or have_only_punct(token)
            or has_numbers_with_punct(token)
            or has_punct_with_numbers(token)
        ):
            return None
        return token

    @cached_property
    def sentences(self) -> List[Span]:
        return list(self._doc.sents)

    @cached_property
    def syllables(self) -> int:
        return sum(len(self._pyphen.positions(str(word))) + 1 for word in self.tokens)

    @cached_property
    def words(self) -> List[str]:
        return [t.text for t in self]

    def unique_words(self) -> Set[str]:
        return set(self.words)

    def characters_per_word(self) -> Dict[str, int]:
        return {word: len(word) for word in self.words}

    def characters(self) -> int:
        return sum([n_char for n_char in self.characters_per_word().values()])

    def filter_words_by_length(self, min_word_length: int = 7) -> List[str]:
        return [word for word in self.words if len(word) >= min_word_length]

    def syllables_per_word(self) -> Dict[str, int]:
        return {word: self._pyphen.positions(word.lower()) for word in self.words}

    def to_df(self):
        return pd.DataFrame(
            [
                [
                    self.words,
                    self.sentences,
                    self.syllables,
                    self.unique_words(),
                    len(self.words),
                    len(self.sentences),
                    self.characters(),
                ]
            ],
            columns=[
                "words",
                "sentences",
                "syllables",
                "unique_words",
                "n_words",
                "n_sentences",
                "n_characters",
            ],
        )

    def __len__(self) -> int:
        return len(self.words)

    def __str__(self) -> str:
        return " ".join(t.text for t in self)

    def __iter__(self) -> Generator[Token, None, None]:
        for token in self.tokens:
            yield token

    def __repr__(self) -> str:
        return f"Document(words={len(self.tokens)}, sentences={len(self.sentences)}, syllables={self.syllables})"



from spacy.tokens.token import Token


def test_sentences(document):
    assert len(document.sentences) == 3


def test_syllables(document):
    assert document.syllables == 204


def test_words(document):
    assert len(document.words) == 80


def test_iter(document):
    for t in document:
        assert isinstance(t, Token)


def test_str(document):
    assert isinstance(str(document), str)


def test_document_repr(document):
    assert repr(document) == "Document(words=80, sentences=3, syllables=204)"
